from collections import defaultdict
from pathlib import Path
from typing import Optional


class FileMatcher:
    def __init__(self, source_files: list[Path]):
        self.source_files = source_files
        self._index: dict[str, Path] = defaultdict()

    def _index_sources(self) -> None:
        for source_file in self.source_files:
            self._index[source_file.stem.lower()] = source_file

    def get_matching_source(self, photo_file: Path) -> Optional[Path]:
        if not self._index:
            self._index_sources()
        return self._index.get(photo_file.stem.lower())



import shutil
from pathlib import Path
from typing import Optional

from rawfinder.exceptions import OverwriteDisabledError
from rawfinder.integrity import FileIntegrityChecker


class FileCopier:
    def __init__(self, overwrite: bool = False, dry_run: bool = False, verifier: Optional[FileIntegrityChecker] = None):
        self.overwrite = overwrite
        self.dry_run = dry_run
        self.verier = verifier

    def copy(self, src: Path, dest_dir: Path) -> bool:
        dest = dest_dir / src.name

        if not self.dry_run:
            dest_dir.mkdir(parents=True, exist_ok=True)
            if dest.exists() and not self.overwrite:
                raise OverwriteDisabledError(str(dest))

        if not self.dry_run:
            self._perform_copy(src, dest)

        if self.verier and not self.dry_run:
            self.verier.verify_copy(src, dest)

        return True

    def _perform_copy(self, src: Path, dest: Path) -> None:
        shutil.copy2(src, dest)



import hashlib
from pathlib import Path

from rawfinder.exceptions import ChecksumError


class FileIntegrityChecker:
    CHUNK_SIZE: int = 4_194_304

    def __init__(self, hash_algorithm: str = "sha256", chunk_size: int = CHUNK_SIZE):
        self.algorithm = hash_algorithm
        self.chunk_size = chunk_size

    def calculate_hash(self, file_path: Path) -> str:
        hash_func = hashlib.new(self.algorithm)
        with open(file_path, "rb") as f:
            while True:
                chunk = f.read(self.chunk_size)
                if not chunk:
                    break
                hash_func.update(chunk)
        return hash_func.hexdigest()

    def verify_copy(self, src: Path, dst: Path) -> None:
        src_hash = self.calculate_hash(src)
        dst_hash = self.calculate_hash(dst)

        if src_hash != dst_hash:
            raise ChecksumError(src, src_hash, dst, dst_hash)



import logging
from pathlib import Path
from typing import Optional

from rawfinder.copier import FileCopier
from rawfinder.exceptions import ChecksumError, DirectoryValidationError, OverwriteDisabledError, UserCancelledError
from rawfinder.finder import FileFinder
from rawfinder.matcher import FileMatcher
from rawfinder.reporters import ProgressReporter

logger = logging.getLogger(__name__)


class RawFinderApp:
    def __init__(
        self,
        photos_dir: Path,
        sources_dir: Path,
        dest_dir: Path,
        copier: FileCopier,
        progress_reporter: ProgressReporter,
        dry_run: bool = False,
        photo_extensions: Optional[list[str]] = None,
        source_extensions: Optional[list[str]] = None,
    ) -> None:
        if source_extensions is None:
            source_extensions = []
        if photo_extensions is None:
            photo_extensions = []
        self.photos_dir = photos_dir
        self.sources_dir = sources_dir
        self.dest_dir = dest_dir
        self.copier = copier
        self.progress_reporter = progress_reporter
        self.dry_run = dry_run
        self.photo_extensions = photo_extensions
        self.source_extensions = source_extensions
        self._validate_directories()

    def _validate_directories(self) -> None:
        for dir_path in [self.photos_dir, self.sources_dir]:
            if not dir_path.is_dir():
                raise DirectoryValidationError(str(dir_path))

    def _find_photo_files(self) -> list[Path]:
        finder = FileFinder(self.photos_dir, set(self.photo_extensions))
        files = finder.find_files()
        logger.info(f"Found {len(files)} photo files in '{self.photos_dir}'")
        return files

    def _find_source_files(self) -> list[Path]:
        finder = FileFinder(self.sources_dir, set(self.source_extensions))
        files = finder.find_files()
        logger.info(f"Found {len(files)} source files in '{self.sources_dir}'")
        return files

    def _match_files(self, photo_files: list[Path], source_files: list[Path]) -> list[tuple[Path, Path]]:
        matcher = FileMatcher(source_files)
        matches = []
        for photo_file in photo_files:
            if source := matcher.get_matching_source(photo_file):
                matches.append((photo_file, source))
            else:
                logger.warning(f"No source match for '{photo_file.name}'")
        return matches

    def _confirm_create_dest_dir(self) -> bool:
        confirm = input(f"Directory '{self.dest_dir}' does not exist. Create it? [Y/n] ")
        return confirm.lower() in ("", "y")

    def _confirm_copy(self, matches: list) -> bool:
        if not matches:
            logger.info("Nothing to copy")
            return False

        if self.dry_run:
            logger.info(f"Dry-run: Would copy {len(matches)} files")
            return True

        confirm = input(f"Copy {len(matches)} files to {self.dest_dir}? [Y/n] ")
        return confirm.lower() in ("", "y")

    def _handle_results(self, results: list[bool]) -> None:
        success_count = sum(1 for r in results if r is True)
        error_count = len(results) - success_count

        logger.info(f"Successfully processed: {success_count} files")
        if error_count > 0:
            logger.warning(f"Failed processing: {error_count} files")

    def _copy_files(self, matches: list[tuple[Path, Path]]) -> None:
        self.progress_reporter.start(len(matches), "Copying RAW files")

        try:
            results = []
            for _, raw_path in matches:
                result = self._process_file(raw_path)
                results.append(result)

            self._handle_results(results)

        finally:
            self.progress_reporter.complete()

    def _process_file(self, raw_path: Path) -> bool:
        try:
            success = self.copier.copy(raw_path, self.dest_dir)
            description = "copied" if success else "failed"
        except ChecksumError as e:
            logger.warning(f"Failed to copy '{raw_path}': {e}")
            description = "failed (checksum mismatch)"
            success = False
        except OverwriteDisabledError:
            logger.warning(f"File '{raw_path}' already exists")
            description = "skipped (overwrite disabled)"
            success = False
        except OSError as e:
            description = "failed"
            success = False
            logger.warning(e)

        self.progress_reporter.update(raw_path.name, success, description)
        return success

    def run(self) -> None:
        photo_files = self._find_photo_files()
        source_files = self._find_source_files()
        matches = self._match_files(photo_files, source_files)

        if not self._confirm_copy(matches):
            raise UserCancelledError

        if not self.dest_dir.exists() and not self._confirm_create_dest_dir():
            raise UserCancelledError

        self.dest_dir.mkdir(parents=True, exist_ok=True)
        self._copy_files(matches)



raw_extensions = [
    # --- Canon ---
    ".crw",  # Old Canon RAW format (before CR2)
    ".cr2",  # Canon RAW 2 (common in DSLRs until ~2018)
    ".cr3",  # Canon RAW 3 (newer DSLRs and mirrorless models)
    # --- Nikon ---
    ".nef",  # Nikon Electronic Format (main Nikon RAW)
    ".nrw",  # Nikon RAW (used by certain Coolpix and compact cameras)
    # --- Sony ---
    ".arw",  # Sony Alpha RAW (current primary format)
    ".srf",  # Sony RAW Format (older models)
    ".sr2",  # Sony RAW 2 (older DSLRs)
    # --- Fujifilm ---
    ".raf",  # Fujifilm RAW (X-series, GFX, etc.)
    # --- Panasonic ---
    ".rw2",  # Panasonic RAW (Lumix series)
    ".raw",  # Some older Panasonic models used ".raw" as extension
    # --- Olympus ---
    ".orf",  # Olympus RAW Format
    # --- Pentax ---
    ".pef",  # Pentax Electronic Format
    # Many Pentax models also support .dng
    # --- Ricoh ---
    # Ricoh acquired Pentax; some Ricoh cameras also produce .dng natively
    # --- Leica ---
    ".rwl",  # Leica RAW (certain lines/models)
    # Leica also heavily uses .dng
    # --- Sigma (Foveon sensors) ---
    ".x3f",  # Sigma RAW (sd, dp series)
    # --- Samsung ---
    ".srw",  # Samsung RAW (NX series)
    # --- Hasselblad ---
    ".3fr",  # Hasselblad RAW (modern format)
    ".fff",  # Older Imacon/Hasselblad format
    # --- Phase One ---
    ".iiq",  # Phase One (digital backs)
    # --- Leaf (part of Phase One) ---
    ".mos",  # Leaf RAW format
    # --- Minolta ---
    ".mrw",  # Minolta RAW (older cameras)
    # --- Kodak (historical DSLRs) ---
    ".dcr",  # Kodak DC RAW
    ".kdc",  # Kodak DC
    ".k25",  # Another historical Kodak RAW format
    # --- Epson ---
    ".erf",  # Epson RAW (R-D1)
    # --- Mamiya ---
    ".mef",  # Mamiya RAW (older digital backs)
    # --- Sinar ---
    ".sti",  # Sinar RAW (rarely encountered)
    # --- ARRI ---
    ".ari",  # ARRI RAW (used in ARRI Alexa and other cinema cameras)
    # --- Smartphones (Apple / Android) & Others: Adobe DNG ---
    ".dng",  # Open "universal" RAW format from Adobe
    # - Apple ProRAW (iPhone) uses .dng
    # - Many Android phones (e.g., Google Pixel) can produce .dng
    # - Leica, Pentax, Ricoh, Hasselblad also use .dng
]



import os
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Any, Optional

import appdirs
import yaml
from pydantic import BaseModel, Field, field_validator

from rawfinder import raw_extensions
from rawfinder.exceptions import InvalidExceptionFormatError, MissingExtensionsError


class ReporterEnum(str, Enum):
    rich = "rich"
    plain = "plain"


@dataclass
class ConfigDefaults:
    verify_checksum: bool = True
    verify_checksum_chunk_size: int = 4 * 1024 * 1024
    reporter: ReporterEnum = ReporterEnum.plain
    dry_run: bool = False
    overwrite: bool = False
    extensions: dict[str, list[str]] = field(
        default_factory=lambda: {
            "photos": [".jpg", ".jpeg"],
            "sources": raw_extensions,
        }
    )


DEFAULTS = ConfigDefaults()


class AppConfig(BaseModel):
    verify_checksum_chunk_size: int = Field(
        default=DEFAULTS.verify_checksum_chunk_size,
        description=f"File verifier chunk size in bytes (default: {DEFAULTS.verify_checksum_chunk_size})",
        ge=1024,
    )
    reporter: ReporterEnum = Field(
        default=DEFAULTS.reporter,
        description=f"Progress reporter type ({[x.value for x in ReporterEnum]}): '{DEFAULTS.reporter}')",
    )
    dry_run: bool = Field(
        default=DEFAULTS.dry_run,
        description=f"Simulate operations without actual changes (default: {DEFAULTS.dry_run})",
    )
    verify_checksum: bool = Field(
        default=DEFAULTS.verify_checksum,
        description=f"Verify file integrity after copy (default: {DEFAULTS.verify_checksum})",
    )
    overwrite: bool = Field(
        default=DEFAULTS.overwrite, description=f"Overwrite existing files (default: {DEFAULTS.overwrite})"
    )
    extensions: dict[str, list[str]] = Field(
        default=DEFAULTS.extensions,
        description="File extensions for different file types (keys: 'photos', 'sources')",
    )

    @field_validator("extensions")
    @classmethod
    def validate_extensions(cls, v: dict[str, list[str]]) -> dict[str, list[str]]:
        required_keys = {"photos", "sources"}
        if not required_keys.issubset(v.keys()):
            missing = required_keys - set(v.keys())
            raise MissingExtensionsError(missing)

        for key, exts in v.items():
            for ext in exts:
                if not ext.startswith("."):
                    raise InvalidExceptionFormatError(ext, key, "Extension '{}' for '{}' must start with a dot (.)")
                if not ext[1:].isalnum():
                    raise InvalidExceptionFormatError(
                        ext, key, "Extension '{}' for '{}' must contain only alphanumeric characters after the dot"
                    )
        return v


class ConfigManager:
    @classmethod
    def get_user_config_path(cls) -> Path:
        config_dir = Path(appdirs.user_config_dir("rawfinder"))
        return config_dir / "config.yaml"

    @classmethod
    def get_default_config_path(cls) -> Path:
        return Path(__file__).parent / "data" / "default_config.yaml"


class ConfigLoader:
    def __init__(self, config_path: Optional[Path] = None):
        self.config_path = config_path

    def load(self) -> AppConfig:
        config_sources = [
            os.environ.get("RAWFINDER_CONFIG"),
            self.config_path,
            ConfigManager.get_user_config_path(),
        ]

        config_data: dict[str, Any] = {}
        for path in config_sources:
            if path and (cfg_path := Path(path)).exists():
                with cfg_path.open() as f:
                    config_data.update(yaml.safe_load(f) or {})
                break
        return AppConfig(**config_data)



import logging
import time
from typing import Optional

from rawfinder.reporters import ProgressReporter


class PlainProgressReporter(ProgressReporter):
    def __init__(self) -> None:
        self.start_time: Optional[float] = None
        self.total = 0
        self.current = 0
        self.logger = logging.getLogger(__name__)

    def start(self, total: int, description: str) -> None:
        self.total = total
        self.current = 0
        self.start_time = time.time()
        self.logger.info(f"{description} [0/{self.total}]")

    def update(self, file: str, success: bool, description: str, advance: int = 1) -> None:
        self.current += advance
        if self.start_time is None:
            raise RuntimeError
        elapsed = time.time() - self.start_time
        self.logger.info(f"[{self.current}/{self.total}] {file} - {description} (Elapsed: {elapsed:.1f}s)")

    def complete(self) -> None:
        if self.start_time is None:
            raise RuntimeError
        elapsed = time.time() - self.start_time
        self.logger.info(f"Completed {self.current}/{self.total} files in {elapsed:.1f} seconds")



from rawfinder.exceptions import UnknownReporterError
from rawfinder.reporters import ProgressReporter
from rawfinder.reporters.plain import PlainProgressReporter
from rawfinder.reporters.rich import RichProgressReporter


class ReporterFactory:
    @staticmethod
    def create(reporter_type: str = "rich") -> ProgressReporter:
        reporters = {
            "plain": PlainProgressReporter,
            "rich": RichProgressReporter,
        }
        try:
            return reporters[reporter_type.lower()]()
        except KeyError as e:
            raise UnknownReporterError(reporter_type) from e



import logging
from pathlib import Path

import click
import yaml

from rawfinder.app import RawFinderApp
from rawfinder.config import AppConfig, ConfigLoader, ConfigManager, ReporterEnum
from rawfinder.copier import FileCopier
from rawfinder.exceptions import (
    BooleanValueExpectedError,
    UnknownReporterError,
    UserCancelledError,
    WriteConfigFileError,
)
from rawfinder.integrity import FileIntegrityChecker
from rawfinder.logging import setup_logging
from rawfinder.reporters import ProgressReporter
from rawfinder.reporters.factories import ReporterFactory

logger = logging.getLogger(__name__)


def str_to_bool(value: str) -> bool:
    if isinstance(value, bool):
        return value
    value = value.lower()
    if value in ("true", "t", "yes", "y", "1"):
        return True
    elif value in ("false", "f", "no", "n", "0"):
        return False
    else:
        raise BooleanValueExpectedError(value)


class NaturalOrderGroup(click.Group):
    def list_commands(self, ctx: click.Context) -> list[str]:
        return list(self.commands.keys())


@click.group(cls=NaturalOrderGroup)
@click.option("--verbose", is_flag=True, help="Enable debug output")
@click.option("--config", type=click.Path(path_type=Path), help="Custom config file path")
@click.option(
    "--handler-type",
    type=click.Choice(["rich"], case_sensitive=False),
    default=None,
    help="Type of logging handler (rich for colorful output, default for standard)",
)
@click.option(
    "--log-file", type=click.Path(path_type=Path), default=None, help="Path to file where logs will be written"
)
@click.pass_context
def cli(ctx: click.Context, verbose: bool, config: Path, handler_type: str, log_file: Path) -> None:
    """RAW Photo Organizer CLI Tool"""
    setup_logging(verbose=verbose, handler_type=handler_type, log_file=log_file)
    ctx.ensure_object(dict)

    cfg_loader = ConfigLoader(config_path=config)
    ctx.obj["config"] = cfg_loader.load()


@cli.command()
@click.argument(
    "photos_dir",
    type=click.Path(
        exists=True,
        file_okay=False,
        path_type=Path,
    ),
)
@click.argument("sources_dir", type=click.Path(exists=True, file_okay=False, path_type=Path))
@click.argument("dest_dir", type=click.Path(path_type=Path))
@click.option("--overwrite", is_flag=True, default=None)
@click.option("--dry-run", is_flag=True, default=None, help="Simulate operations without actual changes")
@click.option(
    "--reporter-type",
    default=None,
    type=click.Choice([item.value for item in ReporterEnum]),
    help="Progress reporting style",
)
@click.option(
    "--verify-checksum",
    type=str_to_bool,
    default=None,
    help="Verify checksums of copied files",
)
@click.option(
    "--verify-checksum-chunk-size",
    type=int,
    default=None,
    help="Chunk size for verifying checksums",
    show_default=True,
)
@click.option(
    "--photos-extensions",
    type=str,
    multiple=True,
    default=None,
    help="Custom photos file extensions (e.g., --photos-extensions .bmp --photos-extensions .png)",
)
@click.option(
    "--sources-extensions",
    type=str,
    default=None,
    multiple=True,
    help="Custom source file extensions (e.g., --sources-extensions .tiff --sources-extensions .pdf)",
)
@click.pass_context
def process(
    ctx: click.Context,
    photos_dir: Path,
    sources_dir: Path,
    dest_dir: Path,
    verify_checksum: bool,
    verify_checksum_chunk_size: int,
    overwrite: bool,
    dry_run: bool,
    reporter_type: str,
    photos_extensions: tuple[str, ...],
    sources_extensions: tuple[str, ...],
) -> None:
    """Organize RAW photos by matching them with JPEG counterparts.

    PHOTOS_DIR - directory with photos (JPEG files)

    SOURCES_DIR - directory with sources files (RAW files)

    DEST_DIR - destination directory for sources (RAW files matched with JPEGs)

    Note: Global options like --verbose, --config, --handler-type, and --log-file can also be used.

    Use `--log-file PATH` to write logs to a file. Run `rawfinder --help` for details on global options.
    """
    config: AppConfig = ctx.obj["config"]

    config.overwrite = overwrite or config.overwrite
    config.dry_run = dry_run or config.dry_run
    config.reporter = ReporterEnum(reporter_type) if reporter_type else config.reporter
    config.verify_checksum = verify_checksum if verify_checksum is not None else config.verify_checksum
    config.verify_checksum_chunk_size = verify_checksum_chunk_size or config.verify_checksum_chunk_size

    if photos_extensions:
        config.extensions["photos"] = list(photos_extensions)
    if sources_extensions:
        config.extensions["sources"] = list(sources_extensions)

    logger.debug("Configuration:")
    logger.debug(yaml.safe_dump(config.model_dump(mode="json"), sort_keys=False))

    try:
        check_sum_verifier = (
            FileIntegrityChecker(chunk_size=config.verify_checksum_chunk_size) if config.verify_checksum else None
        )

        copier = FileCopier(overwrite=config.overwrite, dry_run=config.dry_run, verifier=check_sum_verifier)

        reporter: ProgressReporter = ReporterFactory.create(config.reporter.value)

        app = RawFinderApp(
            photos_dir=photos_dir,
            sources_dir=sources_dir,
            dest_dir=dest_dir,
            copier=copier,
            progress_reporter=reporter,
            dry_run=config.dry_run,
            photo_extensions=config.extensions["photos"],
            source_extensions=config.extensions["sources"],
        )

        app.run()
    except UnknownReporterError as e:
        logger.warning("Invalid reporter type: {}", e.reporter_type)
        raise click.Abort() from e
    except UserCancelledError:
        pass
    except Exception as e:
        logger.exception("Unexpected error")
        raise click.Abort() from e
    else:
        logger.info("Processing completed successfully!")


@cli.group(cls=NaturalOrderGroup)
def config() -> None:
    """Manage application configuration"""
    pass


@config.command()
@click.option("--force", is_flag=True, help="Overwrite existing config")
@click.pass_context
def init(ctx: click.Context, force: bool) -> None:
    """Initialize configuration file"""
    cfg_path = ConfigManager.get_user_config_path()

    if cfg_path.exists() and not force:
        logger.info(f"Config already exists at {cfg_path}")
        return

    try:
        cfg_path.parent.mkdir(parents=True, exist_ok=True)
        with cfg_path.open("w") as f:
            yaml.safe_dump(AppConfig().model_dump(mode="json"), f, sort_keys=False)
    except Exception as e:
        cfg_path.unlink()
        cfg_path.parent.rmdir()
        raise WriteConfigFileError from e

    logger.info(f"Config created at {cfg_path}")


@config.command()
@click.pass_context
def path(ctx: click.Context) -> None:
    """Show config file location"""
    cfg_path = ConfigManager.get_user_config_path()
    logger.info(f"User config: {cfg_path}")
    if not cfg_path.exists():
        logger.warning("Config file does not exist")


@config.command()
@click.pass_context
def edit(ctx: click.Context) -> None:
    """Edit config file in default editor"""
    cfg_path = ConfigManager.get_user_config_path()
    if not cfg_path.exists():
        ctx.invoke(init)

    # Открываем файл в редакторе, используя click.edit()
    updated_content = click.edit(filename=str(cfg_path))
    if updated_content is not None:
        # Если пользователь что-то изменил, перезаписываем файл
        with open(cfg_path, "w") as f:
            f.write(updated_content)
        logger.info("Config updated")
    else:
        logger.info("No changes made to the config")


@config.command()
@click.pass_context
def delete(ctx: click.Context) -> None:
    """Delete config file"""
    cfg_path = ConfigManager.get_user_config_path()
    if cfg_path.exists():
        cfg_path.unlink()
        logger.info("Config deleted")
    else:
        logger.info("Config file does not exist")


@config.command()
@click.pass_context
def validate(ctx: click.Context) -> None:
    """Validate current configuration"""
    try:
        cfg = ConfigLoader().load()
        logger.info("Configuration is valid")
        logger.info(yaml.safe_dump(cfg.model_dump(mode="json"), sort_keys=False))
    except Exception as e:
        logger.info(f"Invalid config: {e!s}")
        raise click.Abort() from e



from rich.progress import BarColumn, Progress, SpinnerColumn, TaskProgressColumn, TextColumn, TimeElapsedColumn

from rawfinder.reporters import ProgressReporter


class RichProgressReporter(ProgressReporter):
    def __init__(self) -> None:
        self._progress = Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            TimeElapsedColumn(),
            TextColumn("[progress.details]{task.fields[details]}"),
            transient=True,
        )

    def start(self, total: int, description: str) -> None:
        self._progress.__enter__()
        self._task_id = self._progress.add_task(description, total=total, details="Initializing...")

    def update(self, file: str, success: bool, description: str, advance: int = 1) -> None:
        self._progress.update(self._task_id, advance=advance, details=description)

    def complete(self) -> None:
        self._progress.__exit__(None, None, None)



import logging
import sys
from pathlib import Path
from typing import Optional

try:
    from rich.logging import RichHandler as _RichHandler

    rich_handler: Optional[type[_RichHandler]] = _RichHandler
except ImportError:
    rich_handler = None


def setup_logging(verbose: bool = False, handler_type: Optional[str] = None, log_file: Optional[Path] = None) -> None:
    level = logging.DEBUG if verbose else logging.INFO
    log_format = "%(asctime)s - %(levelname)s - %(message)s"
    date_format = "%H:%M:%S"

    # Explicitly declare the list as containing logging.Handler objects.
    handlers: list[logging.Handler] = []
    console_handler: logging.Handler = logging.StreamHandler(sys.stdout)

    if handler_type == "rich":
        if rich_handler is None:
            logging.warning("Rich is not installed, using default handler")
            console_handler.setFormatter(logging.Formatter(log_format, datefmt=date_format))
        else:
            console_handler = rich_handler(rich_tracebacks=True)
            console_handler.setFormatter(logging.Formatter("%(message)s"))
    else:
        console_handler.setFormatter(logging.Formatter(log_format, datefmt=date_format))

    handlers.append(console_handler)

    if log_file:
        try:
            log_file.parent.mkdir(parents=True, exist_ok=True)
            file_handler: logging.Handler = logging.FileHandler(log_file)
            file_handler.setFormatter(logging.Formatter(log_format, datefmt=date_format))
            handlers.append(file_handler)
        except Exception as e:
            logging.warning(f"Failed to set up file logging to {log_file}: {e!s}")

    logging.basicConfig(level=level, handlers=handlers, force=True)



from typing import Protocol


class ProgressReporter(Protocol):
    def start(self, total: int, description: str) -> None: ...

    def update(self, file: str, success: bool, description: str, advance: int = 1) -> None: ...

    def complete(self) -> None: ...



import os
from pathlib import Path

from rawfinder.exceptions import FinderDirectoryDoesNotExistError, FinderNoExtensionError


class FileFinder:
    def __init__(self, base_dir: Path, extensions: set[str]):
        if not base_dir.exists():
            raise FinderDirectoryDoesNotExistError(base_dir)
        if not extensions:
            raise FinderNoExtensionError

        self.base_dir = base_dir
        self.extensions = {ext.lower() for ext in extensions}

    def find_files(self) -> list[Path]:
        files = []
        for entry in os.scandir(self.base_dir):
            if entry.is_file() and Path(entry).suffix.lower() in self.extensions:
                files.append(Path(entry))
        return files



import errno
from pathlib import Path
from typing import Optional

from click import ClickException


class FinderDirectoryDoesNotExistError(Exception):
    """Raised when the specified directory does not exist."""

    def __init__(self, base_dir: Path):
        self.base_dir = base_dir
        super().__init__("Directory {} does not exist", str(base_dir))


class FinderNoExtensionError(Exception):
    """Raised when no extensions are specified for the finder."""

    def __init__(self) -> None:
        super().__init__("At least one extension must be specified")


class UnknownReporterError(Exception):
    """Raised when an unsupported reporter type is specified."""

    def __init__(self, reporter_type: str) -> None:
        self.reporter_type = reporter_type
        super().__init__("Reporter type '{}' is not supported", reporter_type)


class DirectoryValidationError(Exception):
    def __init__(self, dir_path: str) -> None:
        super().__init__(f"Directory {dir_path} does not exist")


class UserCancelledError(Exception):
    """Raised when the user cancels the script execution."""

    def __init__(self, message: Optional[str] = None) -> None:
        self.message = message or "User cancelled the script"
        super().__init__(self.message)


class InvalidExceptionFormatError(Exception):
    """Raised when the exception format is invalid."""

    def __init__(self, ext: str, key: str, message: str = "") -> None:
        self.ext = ext
        self.key = key
        super().__init__(message.format(ext, key))


class MissingExtensionsError(ValueError):
    def __init__(self, missing: set[str]) -> None:
        self.missing = missing
        super().__init__(f"Missing extensions for: {missing}")


class OverwriteDisabledError(FileNotFoundError):
    def __init__(self, dest_dir: str) -> None:
        super().__init__(errno.EEXIST, "File {} already exists. Overwrite setting is disabled.", str(dest_dir))


class BooleanValueExpectedError(ClickException):
    def __init__(self, value: str) -> None:
        self.value = value

        super().__init__(f"Boolean value expected (true/false). {value}")


class WriteConfigFileError(ClickException):
    def __init__(self) -> None:
        super().__init__("Error writing config file")


class ChecksumError(ValueError):
    def __init__(self, src: Path, src_hash: str, dst: Path, dst_hash: str) -> None:
        self.src = src
        self.src_hash = src_hash
        self.dst = dst
        self.dst_hash = dst_hash
        super().__init__(f"Checksum mismatch for {src} ({src_hash} != {dst_hash})")
